
-- # import Prop

-- Constants
D     : Set
(<=)  : D -> D -> Prop
(#)   : D -> D -> D
Box   : D -> D

infix 10 ==
infix 12 <=
infix 14 #

(==) = Eq D

-- Axioms

axTrans	  : (x, y, z : D) -> x <= y /\ y <= z  ==>  x <= z
axEq	  : (x, y : D)    -> x == y		 <=>  x <= y /\ y <= x
axAnd	  : (x, y, z : D) -> x <= y # z	 <=>  x <= y /\ x <= z
axMonBox  : (x, y : D)    -> x <= y		 ==>  Box x <= Box y
axBoxAnd  : (x, y : D)    -> Box (x # y) == Box x # Box y

axioms : (x, y, z : D) ->
	     (x <= y /\ y <= z  ==>  x <= z)
	  /\ (x == y  <=>  x <= y /\ y <= x)
	  /\ (x <= y # z  <=>  x <= y /\ x <= z)
	  /\ (x <= y  ==>  Box x <= Box y)
	  /\ (Box (x # y) == Box x # Box y)
axioms x y z = fol-plugin( axTrans x y z
			 , axEq x y
			 , axAnd x y z
			 , axMonBox x y
			 , axBoxAnd x y
			 )
{-
-- Monotone
Mon : (D -> D) -> Prop
Mon f = (x, y : D) -> x <= y  ==>  f x <= f y

-- Fixed points
prefix : (D -> D) -> D -> Prop
prefix f u = f u <= u

postfix : (D -> D) -> D -> Prop
postfix f u = u <= f u

least : (D -> D) -> D -> Prop
least f u = prefix f u /\ ((x : D) -> prefix f x ==> u <= x)

greatest : (D -> D) -> D -> Prop
greatest f u = postfix f u /\ ((x : D) -> postfix f x ==> x <= u)

mu : (D -> D) -> D
nu : (D -> D) -> D

-- Fixed point axioms
muAxiom : (f : D -> D) -> Mon f ==> least f (mu f)
nuAxiom : (f : D -> D) -> Mon f ==> greatest f (nu f)

-- Let's start
sharp : (D -> D) -> D -> D -> D
sharp f q x = q # f x

sharpMon : (f : D -> D) -> (q : D) ->
	    Mon f ==> Mon (sharp f q)
sharpMon = fol-plugin(axioms)

lem1 : (f : D -> D) -> (q : D) ->
	Mon f ==> f (mu (sharp f q)) <= q ==> mu f <= q
lem1 f q monf h = fol-plugin( axioms, rem1
			    , muAxiom g rem1, muAxiom f monf
			    , h
			    )
  where
    g	 = sharp f q

    rem1 : Mon g
    rem1 = sharpMon f q monf

{-
    rem2 : (x : D) -> mu g <= q
    rem2 = fol-plugin(axioms, muAxiom g rem1)

    -- very difficult! ( shouldn't be:: P[a],  Q[a] => _|_,  P[X] => Q[X] )
--     rem3 : (x : D) -> x <= g (mu g)  ==>  x <= mu g
--     rem3 = fol-plugin(axioms, muAxiom g rem1)

    rem3 : (x : D) -> x <= g (mu g)  ==>  x <= mu g
    rem3 = fol-plugin(leqTrans, muAxiom g rem1)

    rem4 : f (mu g) <= g (mu g)
    rem4 = fol-plugin(axioms, h)
-}

AG : D -> D
AG s = nu (sharp Box s)

lem2 : (i, a, s : D) -> i <= a /\ a <= Box a
		       /\ a # s <= Box s /\ i <= s ==>
			  i <= AG s
lem2 i a s h = fol-plugin(axioms, h, rem3)
  where
    rem1 : Mon (sharp Box s)
    rem1 = fol-plugin(axioms, sharpMon)

    rem2 : a # s <= s # Box (a # s)
    rem2 = fol-plugin(h, axioms)

    rem3 : a # s <= AG s
    rem3 = fol-plugin(nuAxiom (sharp Box s) rem1, h, rem2)
-}
