{-# OPTIONS -fglasgow-exts #-}

module TypeChecker.Monad where

import Control.Monad	    ( liftM )
import Control.Monad.State
import Control.Monad.Error
import Control.Monad.Reader
import Data.Set
import Text.PrettyPrint	    ( Doc )

import Interface.Options    ( Options(..) )

import Syntax.Internal
import Syntax.Position	    ( Range )

import TypeChecker.Context
import TypeChecker.UnfoldConstants ( unfoldConstants )
import TypeChecker.Universes
import TypeChecker.Errors

import Utils.Error
import Utils.Map
import Utils.Monads
import Utils.Names

import Utils.Debug

import {-# SOURCE #-} Plugins	-- REMOVE_IF_HADDOCK

{- REMOVE_IF_HADDOCK
-- | This datatype is not really here. It is just an illusion to fool
--   Haddock into accepting recursive modules. The real datatype is
--   'Plugins.SomePlugin'.
data SomePlugin
REMOVE_IF_HADDOCK -}

-- The type checking monad ------------------------------------------------

-- | The type checking monad
type TC = ReaderT Environment (StateT TCState (ErrorT ErrorMsg IO))

-- | The environment used in type checking.
data Environment = Environment
    { envContext	:: Context
	-- ^ The current context. Maps names to types.
    , envErrorMsg	:: ErrorMsg
	-- ^ The current error message. Contains information about
	--   the current activity and position in the source code,
	--   so that should an error occur, we can give a nice message.
    , envOptions	:: Options
    }

emptyEnv :: Environment
emptyEnv = Environment emptyMap noMsg undefined

data TCState = TCState
    { tcUsedNames   :: UsedNames
    , tcPlugins	    :: Map String SomePlugin
    , tcDefinitions :: Definitions
	-- ^ The definitions in scope at the moment.
    }

initState :: TCState
initState = TCState noNamesUsed emptyMap emptyMap

clearState :: TC ()
clearState = modify $ const initState

runTC :: TC a -> IO (Either ErrorMsg a)
runTC tc =
    runErrorT
	$ flip evalStateT initState
	$ flip runReaderT emptyEnv tc

-- | Note: Types as values or terms?
--   Types are values in type checking and we have to check that the
--   type is well-formed before adding it to the environment.
type Definitions = Map Name DefValue

-- Names ------------------------------------------------------------------

instance HasNames TCState where
    getNames	= tcUsedNames
    setNames x	= \st -> st { tcUsedNames = x }

